/* 
 * this file contains the programs for cheking the connectivity of
 * various elements of the circuit, such as which are continous and
 * which are switched capacitors for each capacitor block,
 * and updating the corresponding data structures 
 */

/* 
 * Author: Hormoz Yaghutiel, Advisor: Alberto Sangiovanni-Vincentelli 
 * Henry Chang (henryc@ic.Berkeley.EDU)  - Fall, 1990
 * Last update: March 1992  -  Enrico Malavasi
 */

/* 
 * defined : INPcktCheck INPswLib INPinSw INPpushCap 
 */
#include <stdio.h>
#include "utility.h"
#include "errtrap.h"
#include "ADglobals.h"

#define C_FILE_NAME "INPcktCheck.c"

extern char *cktname;

static struct _cap *ctclist, *swclist;  /* lists of continous and switched caps */
static int          ncontc, nswitc;


static void
  INPinNet(netptr) 
struct _net *netptr;
/* for every sensitive net, find the opamp input terminal.
 * Every capacitor with a node in that net belongs to 
 * the set of continuous capacitors of that integrating block.
 * Every capacitor connected to a switch with a terminal
 * in that net belongs to the set of switched capacitors of that
 * integrating block */
{
    struct _terminal *tmpterm;
    struct _opamp *tmpamp;
    struct _cblk *tmpcblk;
    struct _swch *tmpsw;
    struct _cap *tmpcap;
    struct _terminal *tmpplate;
    
    int input_found;		/* if an opamp input is found */
    
    input_found = NO;
    /*** initialize the stacks */
    ctclist = swclist = NULL;
    ncontc = nswitc = 0;
    
    tmpterm = netptr->firstnett;
    while ( tmpterm != NULL ) {
	if ( tmpterm->capacitor != NULL ) {
	    /*** this belongs to a capacitor */
	    /*** add to the list of continuous capacitors */
	    INPpushCap(tmpterm->capacitor, &ctclist);
	    ncontc++;
	    if ( tmpterm != tmpterm->capacitor->topplate ) {
		/* we have to swap topplate and botplate */
		tmpplate = tmpterm->capacitor->topplate;
		tmpterm->capacitor->topplate = tmpterm->capacitor->botplate;
		tmpterm->capacitor->botplate = tmpplate;
	    }
	} else {
	    if ( tmpterm->module->type == OPAMP ) {
		/* found the opamp input */
		if ( input_found == YES ) {
		    /* more than one opamp associated with the same net */
		    printf("INPinNet: opamps directly connected\n");
		    gerror = ON;
		}
		tmpamp = tmpterm->module->info.opamp;
		if ( tmpterm == tmpamp->inp[MINUS] ) {
		    tmpcblk = tmpamp->cblk[MINUS];
		}
		else if ( tmpterm == tmpamp->inp[PLUS] ) {
		    tmpcblk = tmpamp->cblk[PLUS];
		}
		else {
		    printf(stderr,"INPinNet;1: internal error\n");
		    exit(-1);
		}
		input_found = YES;
	    }
	    else if ( tmpterm->module->type == SWCH ) {
		tmpsw = tmpterm->module->info.swch;
		/* all the capacitors connected to the capterm of this
		   switch belong to the set of switched capacitor of this integ */
		INPinSw(tmpsw);
	    }
	    else if ( tmpterm->module->type != CBLK ) {
		/*** internal error */
		fprintf(stderr,"INPinNet;1: internal error\n");
		exit(-1);
	    }
	}
	tmpterm = tmpterm->nxtnett;
    }
    if ( input_found == NO ) {
	printf("INPprocessNet;2: internal error\n");
	exit(-1);
    }
    tmpcblk->ncontc = ncontc;
    tmpcblk->nswitc = nswitc;
    
    tmpcblk->contcaps = ctclist;
    tmpcblk->swcaps = swclist;
    
    /* now for each capacitor belonging to this block update the pointers */
    tmpcap = tmpcblk->contcaps;
    while ( tmpcap != NULL ) {
	tmpcap->type = CONT;
	tmpcap->cblk = tmpcblk;
	
	tmpcap->topplate->channel = chanlist[BOTTOM];
	tmpcap->topplate->module = tmpcblk->module;
	tmpcap->topplate->chrlayer = 2;
	tmpcap->topplate->chloc = TOP;
	
	tmpcap->botplate->channel = chanlist[BOTTOM];
	tmpcap->botplate->module = tmpcblk->module;
	tmpcap->botplate->chrlayer = 1;
	tmpcap->botplate->chloc = TOP;
	
	if ( tmpcap->botplate->netptr->module == tmpcblk->opamp->module ) {
	    /* this is an integrating capacitor */
	    if ( tmpcblk->integcap != NULL ) {
		printf("INPinNet: more than one integrating cap per input\n");
		gerror = ON;
	    }
	    tmpcblk->integcap = tmpcap;
	}
	tmpcap = tmpcap->next;
    }
    tmpcap = tmpcblk->swcaps;
    while( tmpcap != NULL ) {
	tmpcap->type = SWITCHED;
	tmpcap->cblk = tmpcblk;
	tmpcap->topplate->channel = chanlist[TOP];
	tmpcap->topplate->module = tmpcblk->module;
	tmpcap->topplate->chrlayer = 2;
	tmpcap->topplate->chloc = BOTTOM;
	
	tmpcap->botplate->channel = chanlist[TOP];
	tmpcap->botplate->module = tmpcblk->module;
	tmpcap->botplate->chrlayer = 1;
	tmpcap->botplate->chloc = BOTTOM;
	
	tmpcap = tmpcap->next;
    }
}

static void
  INPswType(lib)
char *lib;				/* directory name for library */
/* Find type of each switch and sensitivity of their capterm nets
 * add them to their respective library lists */
{
    struct _module *tmpmod;
    struct _swch *tmpsw;
    struct _net *netptr1, *netptr2, *netptr;


    /* go over the list of all switches */
    tmpmod = modlist[SWCH];
    while ( tmpmod != NULL ) {
	tmpsw = tmpmod->info.swch;
	netptr1 = tmpsw->amp1term->netptr;
	netptr2 = tmpsw->amp2term->netptr;
	if ( netptr1->type == GROUND && netptr2->type == GROUND ) {
	    printf("INPswType: both end grounded switch\n");
	    gerror = ON;
	}
	else if (netptr1->type != GROUND && netptr2->type != GROUND ) {
	    if ( netptr1->sensitive == YES || netptr2->sensitive == YES ) {
		printf("INPcktCheck: wrong switch type for input\n");
		gerror = ON;
	    }
	    tmpsw->type = SWOUT3;
	}
	else {
	    /*** only one ampnode is not grounded */
	    if ( netptr1->type != GROUND ) {
		netptr = netptr1;
	    }
	    else {
		netptr = netptr2;
	    }
	    if ( netptr->sensitive == YES ) {
		tmpsw->type = SWIN;
		/*** this switch is connected to the input of an opamp */
		/*** the sensitivity is passed on to capterm */
		tmpsw->capterm->netptr->sensitive = YES;
	    }
	    else {
		tmpsw->type = SWOUT2;
	    }
	}
	tmpmod = tmpmod->next;
    }
    
    /*** add the complement clock terminals */
    INPswClkBarT();
    
    /*** now add the switches to their respective library lists */
    INPswLib(lib);
}

void
  INPcktCheck(lib)
char *lib;				/* directory name for library */
{
    struct _net *tmpnet;
    struct _module *tmpmod;
    struct _terminal *tmpterm;

    for(tmpnet=netlist ; tmpnet!=NULL ; tmpnet=tmpnet->next) {
	/*  Count the number of CLOCK nets  */
	if( tmpnet->type == CLOCK )    Numclk++;
	if( tmpnet->type == IN )       INPinNet(tmpnet);		
    }

    INPswType(lib);

    /* Find if any capacitor is not associated with any integrator */
    tmpnet = netlist;
    while ( tmpnet != NULL ) {
	tmpterm = tmpnet->firstnett;
	while ( tmpterm != NULL ) {
	    if ( tmpterm->capacitor != NULL && tmpterm->module == NULL ) {
		printf("*** ERROR: capacitor %s belongs to no opamps\n",
		       tmpterm->capacitor->name);
		gerror = ON;
	    }
	    tmpterm = tmpterm->nxtnett;
	}
	tmpnet = tmpnet->next;
    }
}


/* 
 * For each switch of the type !SWIN allocate memory for clock complement
 * terminals 
 */
INPswClkBarT()
{
  /* char *netname, tmpname[60];  old yucky stuff -hcc */
  char *netname;
  struct _module *tmpmod;
  struct _swch *tmpsw;
  struct _net *tmpnet, *INPtermInsert();

/* What does declaring this like this mean */

  struct _terminal *INPcrtTerm();

  for ( tmpmod = modlist[SWCH]; tmpmod != NULL; tmpmod = tmpmod->next ) {
    tmpsw = tmpmod->info.swch;
    MALLOC_ERRCHK (netname, char, 60);
    if ( tmpsw->type != SWIN ) {
      sprintf( netname, "%s%s", tmpsw->clk1term->netptr->name, "_bar" );

    /* 
     * Fails if there is a netname already of this in the linked list.
     * Attempts to FREE, netname, wasn't originally allocated by malloc
     *
     * on mips machines (and on vaxes)
     * adore2: fatal error detected by mm (code -3):
     *         attempt to free() bogus pointer
     * -hcc
     *
     * Checking original program, can't believe that this had worked.
     * If this didn't cause it to crash, then this symbol table that
     * is made was not used.
     * Conclusion: free() was used on bogus pointers in original
     * program.  mm caught the error which wasn't included in the
     * original program.  Who knows????			-hcc
     */

      tmpnet = INPtermInsert( &netname );

      tmpnet->type = CLOCK;
      tmpsw->clk1bart = INPcrtTerm(tmpnet, chanlist[CLKCH], CLOCK,
				   tmpmod, NULL);
      MALLOC_ERRCHK (netname, char, 60);
      sprintf( netname, "%s%s", tmpsw->clk2term->netptr->name, "_bar" );
      tmpnet = INPtermInsert( &netname );
      tmpnet->type = CLOCK;
      tmpsw->clk2bart = INPcrtTerm(tmpnet, chanlist[CLKCH], CLOCK,
				   tmpmod, NULL);
    }
  }
} /* INPswClkBarT */

/* 
 * create library name swin, swout2 and swout3, and add the switches 
 *
 * This is where the library OCT instances are stored.
 * They were orignally in sprintf(tmptk, "%s%s", cktname, "/SWIN"); 
 * Change to new format.
 * -hcc
 */
INPswLib(lib) 
char *lib;				/* directory name for library */
{
  struct _module *tmpmod;
  struct _swch *tmpsw;
  struct _lib *INPlibInsert(), *tmplib;
  /* 
   * char *token, tmptk[100];
   * This is bogus, dude!	-hcc 
   */
  char *token;

  /*
   * This routine has the same problem as INPswClkBarT() does.
   * If a duplicate token appears, then
   *
   * adore2: fatal error detected by mm (code -3):
   *     attempt to free() bogus pointer
   *
   * occurs.
   * 
   * Will fix in the same manner as above.
   */

/* Completely restructure.
  for ( tmpmod = modlist[SWCH] ; tmpmod != NULL; tmpmod = tmpmod->next ) {
    if ( tmpmod->libptr == NULL ) {
      tmpsw = tmpmod->info.swch;
      if ( tmpsw->type == SWIN ) {
        MALLOC_ERRCHK (tmptk, char, 100);
	sprintf(tmptk, "%s%s", cktname, "/SWIN");
	token = tmptk;
	/*
	  token = "../lib/SWIN";
	  **
	tmplib = INPlibInsert(&token);
	INPaddLib( tmpmod, tmplib );
      } else if ( tmpsw->type == SWOUT2 ) {
        MALLOC_ERRCHK (tmptk, char, 100);
	sprintf(tmptk, "%s%s", cktname, "/SWOUT2");
	token = tmptk;
	/*
	  token = "../lib/SWOUT2";
	  **
	tmplib = INPlibInsert(&token);
	INPaddLib( tmpmod, tmplib );
      } else if ( tmpsw->type == SWOUT3 ) {
        MALLOC_ERRCHK (tmptk, char, 100);
	sprintf(tmptk, "%s%s", cktname, "/SWOUT3");
	token = tmptk;
	/*
	  token = "../lib/SWOUT3";
	  **
	tmplib = INPlibInsert(&token);
	INPaddLib( tmpmod, tmplib );
      } else {
	printf(stderr,"INPswLib;1: internal error\n");
	exit(-1);
      }
    }
  }
*/ 
    for ( tmpmod = modlist[SWCH] ; tmpmod != NULL; tmpmod = tmpmod->next ) {
        if ( tmpmod->libptr == NULL ) {
            tmpsw = tmpmod->info.swch;
            MALLOC_ERRCHK (token, char, 100);
            strcpy (token, lib); 
            switch (tmpsw->type) {
                case SWIN :
                    strcat (token, "/SWIN");
                    break;
                case SWOUT2 :
                    strcat (token, "/SWOUT2");
                    break;
                case SWOUT3 :
                    strcat (token, "/SWOUT3");
                    break;
                default:
                    errRaise (C_FILE_NAME,0,"INPswLib;1: internal error");
            } /* switch */
            tmplib = INPlibInsert (&token);
            INPaddLib (tmpmod, tmplib);
        } /* if */
    } /* for */
}


INPinSw(swch)
struct _swch *swch;
{
  struct _net *tmpnet;
  struct _terminal *tmpterm;
  struct _terminal *tmpplate;

  tmpnet = swch->capterm->netptr;

  tmpterm = tmpnet->firstnett;
  while ( tmpterm != NULL ) {
    if( tmpterm->capacitor != NULL ) {
      /*** this terminal belongs to a capacitor */
      /*** add the capacitor to the list of switched capacitors */
      INPpushCap(tmpterm->capacitor, &swclist);
      nswitc++;
      if ( tmpterm != tmpterm->capacitor->topplate ) {
	/* we have to swap topplate and botplate */
	tmpplate = tmpterm->capacitor->topplate;
	tmpterm->capacitor->topplate = tmpterm->capacitor->botplate;
	tmpterm->capacitor->botplate = tmpplate;
      }
    }
    tmpterm = tmpterm->nxtnett;
  }
}

INPpushCap(capacitor, pcapstack) 
struct _cap *capacitor; /* to be pushed in the stack */
struct _cap **pcapstack; /* the stack */
{
  if ( capacitor == NULL ) {
    /*** internal error */
    printf(stderr,"INPpushCap;1: internal error\n");
    exit(-1);
  }
  capacitor->next = *pcapstack;
  *pcapstack = capacitor;
}

