/*
 *  set.h -- definitions for packed arrays of bits
 */


typedef struct set_struct set_t;
struct set_struct {
    int size;
    int nwords;
    unsigned data[1];
};


typedef enum set_error_enum {
    SET_SUBSCRIPT,
    SET_SIZE_MISMATCH,
    SET_NWORDS_MISMATCH,
    SET_NWORDS_BAD,
    SET_DIRTY
} set_error_t;


/* Define host machine size of "unsigned int" */
#ifndef BPI
#define BPI             32              /* # bits per integer */
#endif

#if BPI == 32
#define LOGBPI          5               /* log(BPI)/log(2) */
#else
#define LOGBPI          4               /* log(BPI)/log(2) */
#endif

/*
 *  Allow for debugging
 */
#ifdef DEBUG
#define CHECK(a, e)		\
    ((e) >= a->size || (e) < 0) ? set_error(SET_SUBSCRIPT) :
#define DEBUG1(a)		set_check1(a)
#define DEBUG2(a, b)		set_check2(a, b)
#define DEBUG3(a, b, c)		set_check3(a, b, c)
#else
#define CHECK(a, e)
#define DEBUG1(a)
#define DEBUG2(a, b)
#define DEBUG3(a, b, c)
#endif

#define set_free(r)		(free((char *) r), r = 0)

#define WHICH_WORD(element)	((element) >> LOGBPI)
#define WHICH_BIT(element)	((element) & (BPI-1))

#define is_in_set(s, e)		\
    (CHECK(s, e) (s->data[WHICH_WORD(e)] & (1 << WHICH_BIT(e))))
#define set_remove(s, e)	\
    (CHECK(s, e) (s->data[WHICH_WORD(e)] &= ~(1 << WHICH_BIT(e))))
#define set_insert(s, e)	\
    (CHECK(s, e) (s->data[WHICH_WORD(e)] |= 1 << WHICH_BIT(e)))



/*
 *  in-line macros
 */
#define ODD_WORD(r)	\
    ((~ (unsigned) 0) >> (r->nwords-1) * BPI - r->size)

#define SET_CLEAR(r) { 		\
    register int _i;		\
    for(_i = r->nwords-1; _i >= 0; _i--) r->data[_i] = 0;\
}

#define SET_FILL(r) {		\
    register int _i;		\
    for(_i = r->nwords-2; _i >= 0; _i--) r->data[_i] = ~0;\
    r->data[r->nwords-1] = ODD_WORD(r);\
}

#define SET_COPY(r, a) {	\
    register int _i;		\
    DEBUG2(r, a);		\
    for(_i = a->nwords-1; _i >= 0; _i--)\
	r->data[_i] = a->data[_i];\
}

#define SET_AND(r, a, b) {	\
    register int _i;		\
    DEBUG3(r, a, b);		\
    for(_i = a->nwords-1; _i >= 0; _i--)\
	r->data[_i] = a->data[_i] & b->data[_i];\
}

#define SET_OR(r, a, b) {	\
    register int _i;		\
    DEBUG3(r, a, b);		\
    for(_i = a->nwords-1; _i >= 0; _i--)\
	r->data[_i] = a->data[_i] | b->data[_i];\
}

#define SET_DIFF(r, a, b) {	\
    register int _i;		\
    DEBUG3(r, a, b);		\
    for(_i = a->nwords-1; _i >= 0; _i--)\
	r->data[_i] = a->data[_i] & ~ b->data[_i];\
}

#define SET_COMPARE(a, b, x) {	\
    register int _i;		\
    DEBUG2(a, b);		\
    x=0; for(_i = a->nwords-1; _i >= 0; _i--)\
	if (a->data[_i] > b->data[_i]) {x = -1; break;}\
	else if (a->data[_i] < b->data[_i]) {x = 1; break;}\
}

#define SETP_EMPTY(a, x) {	\
    register int _i;		\
    DEBUG1(a);			\
    x=1; for(_i = a->nwords-1; _i >= 0; _i--)\
	if (a->data[_i]) {x=0; break;}\
}

#define SETP_EQUAL(a, b, x) {	\
    register int _i;		\
    DEBUG2(a, b);		\
    x=1; for(_i = a->nwords-1; _i >= 0; _i--)\
	if (a->data[_i] != b->data[_i]) {x=0; break;}\
}

#define SETP_DISJOINT(a, b, x) {\
    register int _i;		\
    DEBUG2(a, b);		\
    x=1; for(_i = a->nwords-1; _i >= 0; _i--)\
	if (a->data[_i] & b->data[_i]) {x=0; break;}\
}

#define SETP_IMPLIES(a, b, x) {	\
    register int _i;		\
    DEBUG2(a, b);		\
    x=1; for(_i = a->nwords-1; _i >= 0; _i--)\
	if (a->data[_i] & ~ b->data[_i]) {x=0; break;}\
}


extern void free();
extern void abort();

extern void set_clear();
extern void set_fill();
extern void set_copy();
extern void set_and();
extern void set_or();
extern void set_diff();
extern void set_check1();
extern void set_check2();
extern void set_check3();
extern set_t *set_new();
extern set_t *set_new_copy();
extern int setp_empty();
extern int setp_full();
extern int setp_disjoint();
extern int set_ord();
extern int set_error();
extern char *set_format();
extern char *set_format_bv();


/*
 *  this makes dbx happy about the types when debugging
 */
#ifndef lint
static void set_debug(s)
set_t *s;
{
    printf("set is %s\n", set_format(s));
    printf("set is %s\n", set_format_bv(s));
}
#endif
